# Diffusion-Limited Aggregation in 2-d
# originally written in FORTRAN
# 2024, Yoshinori Hayakawa

import numpy as np
import random
import math

LSIZ=513 
NCLUSTER=100000
NDIM=2 
CGF=2 
LAYER=0.2 
MF=10 
OF=15 
NOFFSET=10 

class Dla2d:
    noa=0 
    r_tip=0.0
    r_kill=0.0

    def __init__(self):
        self.init_vars()
        self.set_seed()

    def init_vars(self):
        self.u = np.zeros(shape=(LSIZ,LSIZ),dtype=np.uint32)
        self.pp = np.zeros(shape=(NCLUSTER,),dtype=np.uint32)
        self.x_cluster = np.zeros(shape=(NDIM,NCLUSTER))
        self.x_walker = np.zeros(shape=(NDIM,))
        self.x_cluster[:,0] = (LSIZ-1)*CGF/2 

    def set_seed(self):
        self.r_tip = 0.0
        self.r_kill = OF
        self.noa = 0
        self.x_walker = self.x_cluster[:,0]
        self.grow()

    def step_fwd(self):
        while True:
            rr = 0.0
            u0 = np.random.normal(0,1,size=(NDIM,))
            rr = np.sqrt(np.sum(u0**2))
            u0 = u0 / rr
            self.x_walker = (self.r_tip + 5.0) * u0 + self.x_cluster[:,0]
            if self.random_walk() > 0:
                break ;  # random_walk == 0 : fail 

    def get_dist(self):
        index = 0
        dd = np.linalg.norm(self.x_walker - self.x_cluster[:,0])
        if dd > self.r_tip+5.0:
            distance = dd - self.r_tip - LAYER 
            return (index, distance)

        step = 5
        while True:
            ibd = (1 << step) * CGF
            ix = self.x_walker.astype(int) // ibd
            if (self.u[ix[0],ix[1]] & (1<<step)) == 0:
                distance = (2**step) * CGF - 1.0
                return (index, distance)
            step = step - 1
            if step<0:
                break

        ix = self.x_walker.astype(int) // CGF 
        dd = 1.0e+30
        for i1 in range(ix[1]-1, ix[1]+2, 1):
            for i0 in range(ix[0]-1, ix[0]+2, 1):
                ell = self.u[i0,i1] >> NOFFSET
                while True:
                    if ell==0:
                        break 
                    d = np.sum((self.x_cluster[:,ell] - self.x_walker)**2)
                    if d < dd:
                        dd = d
                        index = ell
                    ell = self.pp[ell] 

        dd = math.sqrt(dd) - 1.0   #  1 = diameter of paticle
        if dd > CGF:
            distance = CGF - 1.0 
        else:
            distance = dd
        return (index,distance)


    def random_walk(self):
        while True:
            index,distance = self.get_dist()
            if distance > LAYER: # large leap 
                dx = np.random.normal(0,1,size=(NDIM,))
                rr = np.sqrt(np.sum(dx**2))
                dx = distance * dx / rr
            else: # touching the surface 
                dx = self.x_walker - self.x_cluster[:,index]
                rr = np.sqrt(np.sum(dx**2))
                dx = dx / rr * (1.0 - rr) # adjust the location 
                self.x_walker = self.x_walker + dx
                return 1 

            self.x_walker = self.x_walker + dx 
            rr = np.linalg.norm(self.x_walker - self.x_cluster[:,0]) # check how far from the origin
            if rr > self.r_kill:
                return 0 

    def grow(self):
        self.noa += 1
        
        self.x_cluster[:,self.noa] = self.x_walker[:] 
        dd = np.linalg.norm(self.x_walker - self.x_cluster[:,0])
        if dd > self.r_tip:
            self.r_tip = dd 
        self.r_kill = MF * self.r_tip + OF 

        ix = self.x_walker.astype(int) // CGF 
        k = (self.u[ix[0],ix[1]]) >> NOFFSET
    
        if k==0:
            self.u[ix[0],ix[1]] |= (self.noa * (1<<NOFFSET))
            self.pp[self.noa] = 0
        else:
            ell = self.pp[k]
            while True:
                if (ell==0):
                    break 
                k = ell
                ell = self.pp[k]
            self.pp[k] = self.noa
            self.pp[self.noa] = 0 

        step=0
        while step<8:
            ibd = (1<<step) * CGF
            ix = self.x_walker.astype(int) // ibd 
            for i1 in range(ix[1]-1,ix[1]+2,1):
                for i0 in range(ix[0]-1,ix[0]+2,1):
                    if i0>=0 and i0<LSIZ and i1>=0 and i1<LSIZ:
                        self.u[i0,i1] |= (1<<step)
            step += 1

        return

    def print(self):
        print(self.x_cluster[0,self.noa], self.x_cluster[1,self.noa])

###########
import pyglet

LSIZ=513 
CGF=2
WSIZ=LSIZ*CGF

dla = Dla2d()

N=20000

config = pyglet.gl.Config(double_buffer=True)
win = pyglet.window.Window(WSIZ,WSIZ,config=config,resizable=False)
win.set_caption('Diffusion-limited Aggregation')

circles = [ ]

@win.event
def on_draw():
    win.clear()
    for c in circles:
        c.draw()
    n = len(circles)
    label = pyglet.text.Label('N='+str(n), font_name='Helvetica', font_size=24, color=(128,128,128,255), \
                              x=10, y=10, anchor_x='left', anchor_y='bottom')
    label.draw()
    return

def update(dt):
    n = dla.noa
    if n < N:
        dla.step_fwd()
        dla.grow()
        cluster = dla.x_cluster        
        x = (cluster[0,n]-WSIZ/2)*4 + WSIZ/2
        y = (cluster[1,n]-WSIZ/2)*4 + WSIZ/2
        circle = pyglet.shapes.Circle(x=x, y=y, radius=2, color=(50, 100, 200))
        circles.append(circle)

pyglet.clock.schedule(update)
pyglet.app.run()
